<?php

/**
 * @file
 * Entity Lister class file.
 */

class EntityLister {

  public $config;
  public $delta;
  public $flag;
  public $hostItem;

  /**
   * Construct the list object.
   *
   * @param array $config
   *   The config options for the list, in an associative array.  See below.
   * @param string $delta
   *   The delta.
   * @param string $flag
   *   An identifier.  When this class is used for a field formatter, this will
   *   be a field_name.  In cases where the field is configured to support
   *   multiple values, all values in the field should have the same string set
   *   for the $flag, but each should have a different $delta.
   * @param array $host_item
   *   When this class is used for a field formatter, this array describes the
   *   item to which the field is attached. See below.
   *
   * The $config array:
   *   $config['type']
   *   $config['bundle']
   *   $config['view_mode']
   *   $config['numitems']
   *   $config['sort']
   *   $config['direction']
   *   $config['terms']
   *   $config['pager']
   *   $config['cache']
   *   $config['status']
   *
   * The $host_item array:
   *   $host_item[0] The host item's entity_id.
   *   $host_item[1] The host item's revision_id.
   *   $host_item[2] The host item's bundle.
   *   $host_item[3] The host item's entity type.
   */
  public function __construct($config = array(), $delta = '0', $flag = '', $host_item = array()) {

    $this->config = $config;
    $this->delta = $delta;
    $this->flag = $flag;
    $this->hostItem = $host_item;

  }

  /**
   * For the standard pager, determine what page we're on.
   *
   * @return string
   *   The page number, e.g. '2'.
   */
  public function getPageNum() {

    return isset($_GET['page']) ? $_GET['page'] : '';

  }

  /**
   * Return the list as themed output.
   *
   * @param string $page
   *   The page we're on. For paged results, this will be '0' or an empty
   *   string for the first page of results, '1' for the second page, and so on.
   *
   * @return string
   *   The themed output.
   */
  public function getList($page = '') {

    /*
     * $this->config contains the end user's "settings" for this list,
     * e.g. what the user defines in the entity's create/edit forms.
     */
    $c = $this->config;

    $config['type'] = isset($c['type']) ? $c['type'] : 'node';
    $config['bundle'] = isset($c['bundle']) ? $c['bundle'] : json_encode(array('article'));
    $config['view_mode'] = isset($c['view_mode']) ? $c['view_mode'] : 'teaser';
    $config['numitems'] = isset($c['numitems']) ? $c['numitems'] : '3';
    $config['sort'] = isset($c['sort']) ? $c['sort'] : 'created';
    $config['direction'] = isset($c['direction']) ? $c['direction'] : 'DESC';
    $config['terms'] = isset($c['terms']) ? $c['terms'] : '';
    $config['pager'] = isset($c['pager']) ? $c['pager'] : '2';
    $config['cache'] = isset($c['cache']) ? $c['cache'] : '1';
    $config['pager_position'] = isset($c['pager_position']) ? $c['pager_position'] : '1';
    $config['status'] = isset($c['status']) ? $c['status'] : 1;

    if ('0' === $config['numitems']) {
      return '';
    }

    extract($config);

    $bundles = json_decode($bundle);

    // Check if the user can view the items.
    $access = $this->accessCheck($type, $bundles);
    if (!$access[0]) {
      return '';
    }
    elseif ('node' == $type) {
      // Bundles the user can view.
      $bundles = $access[1];
    }

    // Always require one or more bundles.
    if (empty($bundles)) {
      return '';
    }

    if ('1' === $cache) {
      // Include the cache-related functions.
      module_load_include('inc', 'entity_lister', 'entity_lister.cache');
      // Set a cache ID based on the config.
      $cid['type'] = $type;
      $cid['bundle'] = implode(',', $bundles);
      $cid['view_mode'] = $view_mode;
      $cid['numitems'] = $numitems;
      $cid['sort'] = $sort;
      $cid['direction'] = $direction;
      $cid['terms'] = $terms;
      $cid['pager'] = $pager;
      $cid['pager_position'] = $pager_position;
      $cid['status'] = $status;
      // We also need the page for uniqueness.
      $cid['page'] = $page;
      // Add the language.
      global $language;
      $cid['language'] = $language->language;
      $cid = 'entity_lister:' . implode(':', $cid);
      // Return the cached data if we have any.
      if ($cachedata = cache_get($cid)) {
        if ('2' === $pager) {
          $this->pager();
        }
        return $cachedata->data;
      }
    }

    /* Tell EntityFieldQuery and the entity's view function what page
     * we're on. We can set this any time before we execute the query.
     */
    $_GET['page'] = $page;

    $query = new EntityFieldQuery();
    $query->entityCondition('entity_type', $type);

    // Don't display the host entity in its own list.
    if (!empty($this->hostItem)) {
      if ($type == $this->hostItem[3]) {
        if (!isset($bundles) || in_array($this->hostItem[2], $bundles)) {
          $query->entityCondition('entity_id', $this->hostItem[0], '!=');
        }
      }
    }

    $prop_info = entity_get_property_info($type);
    $props = $prop_info['properties'];

    // Deal with lingering inconsistencies re: how entities store bundle data.
    // Checking the type property works seemingly for everything except
    // taxonomy_term bundles.
    if (isset($props['type'])) {
      $query->propertyCondition('type', $bundles);
    }
    // This entityCondition works in some cases, taxonomy_term being one
    // of those cases.
    elseif ('taxonomy_term' == $type) {
      $query->entityCondition('bundle', $bundles);
    }

    // Not all entity types have a status property.
    if (isset($props['status'])) {
      $query->propertyCondition('status', $status);
    }

    // Not all entity types have a sort property.
    if (isset($props[$sort])) {
      $query->propertyOrderBy($sort, $direction);
    }

    $bundle_fields = array();
    foreach ($bundles as $bundle) {
      $bundle_fields += field_info_instances($type, $bundle);
    }
    // We use $bundle_fields here and in the call to safeLoop().
    $query = $this->taxonomy($query, $terms, $bundle_fields);

    if ('0' !== $pager) {
      $query->pager($numitems);
    }
    else {
      $query->range(0, $numitems);
    }

    $result = $query->execute();

    if (!$result || !isset($result[$type])) {
      return '';
    }
    else {
      $ids = array_keys($result[$type]);
      $entities = entity_load($type, $ids);
    }

    if ($entities) {
      $entities = $this->safeLoop($entities, $config, $bundle_fields);
      $output = $this->theme($entities, $config, $page);
    }
    else {
      $output = '';
    }

    if ('1' === $cache) {
      cache_set($cid, $output, 'cache', CACHE_TEMPORARY);
    }

    return $output;

  }

  /**
   * Wrap the list in an element for rendering.
   *
   * @param string $list
   *   The themed list output.
   *
   * @return array
   *   The element.
   */
  public function element($list) {

    if (!empty($this->flag)) {
      $f = str_replace('_', '-', $this->flag) . '-';
    }
    else {
      $f = '';
    }

    // The AJAX pager uses this id.
    $id = 'entity-list-' . check_plain($f . $this->delta);

    $element[$this->delta] = array(
      '#prefix' => '<div id="' . $id . '">',
      '#suffix' => '</div>',
      '#markup' => $list,
    );

    return $element;

  }

  /**
   * Theme the list.
   *
   * @param array $entities
   *   The entities to list.
   * @param string $config
   *   The list config as defined in the widget form.
   *
   * @return string
   *   The themed output.
   */
  protected function theme($entities, $config) {

    extract($config);

    if ('2' === $pager) {
      $this->pager();
    }
    if ('0' !== $pager) {
      $above = '0' === $pager_position || '2' == $pager_position;
      $below = '1' === $pager_position || '2' == $pager_position;
    }
    else {
      $above = $below = FALSE;
    }

    $items = $this->renderItems($entities, $type, $view_mode);

    $output = '';
    if ($above) {
      $output .= theme('pager');
    }
    $output .= theme('item_list', array('items' => $items));
    if ($below) {
      $output .= theme('pager');
    }

    return $output;
  }

  /**
   * Render list items.
   *
   * @param array $entities
   *   The entities to list.
   * @param string $type
   *   The entity type.
   * @param string $view_mode
   *   The view mode.
   *
   * @return array
   *   The rendered items.
   */
  protected function renderItems($entities, $type, $view_mode) {

    $viewfunc = $type . '_view';

    if (function_exists($viewfunc)) {
      foreach ($entities as $entity) {
        $items[] = drupal_render($viewfunc($entity, $view_mode));
      }
    }
    elseif (function_exists('theme_entity_lister_' . $type)) {
      // How to provide a custom theme function in cases when an entity type
      // has no view function, as is the case with the taxonomy_vocabulary type.
      // See theme_entity_lister_taxonomy_vocabulary() for an example.
      $vars['entities'] = $entities;
      $items = theme('entity_lister_' . $type, $vars);
    }
    else {
      $msg = 'No view function or theme function in EntityLister::renderItems().';
      watchdog('entity_lister', $msg);
      // No view function and no theme function, so just list the titles.
      foreach ($entities as $entity) {
        $items[] = $entity->name;
      }
    }

    return $items;
  }

  /**
   * Activate the ajax pager.
   */
  protected function pager() {

    $path = drupal_get_path('module', 'entity_lister');
    drupal_add_js($path . '/ajax-pager.js');

    $js_settings[$this->delta] = array(
      'config' => $this->config,
      'delta' => $this->delta,
      'flag' => $this->flag,
      'host_item' => $this->hostItem,
      'list_class' => get_class($this),
    );
    drupal_add_js(array('entity_lister' => $js_settings), 'setting');

  }

  /**
   * Check if the user can view the list items.
   *
   * @param array $type
   *   The entity type.
   * @param string $bundles
   *   The bundles.
   *
   * @return array
   *   Info about the user's access to this type and its bundles.
   *   $info[0] bool -- Can we view the type?
   *   $info[1] array -- For nodes, the bundles we can view.
   */
  protected function accessCheck($type, $bundles) {

    $info = array(TRUE, array());

    if ('node' == $type) {

      if (!user_access('access content')) {
        $info[0] = FALSE;
      }
      else {
        // Add the node_view_permissions integration.
        $bundles = $this->nodeViewPermissions($bundles);
        $info[1] = $bundles;
      }

    }
    elseif ('user' == $type) {
      if (!user_access('access user profiles')) {
        $info[0] = FALSE;
      };
    }
    elseif ('comment' == $type) {
      if (!user_access('access comments')) {
        $info[0] = FALSE;
      };
    }
    return $info;
  }

  /**
   * Taxonomy integration.
   *
   * Add a fieldCondition to query only for items that have been assigned
   * one of the terms.
   *
   * @param object $query
   *   The query.
   * @param string $terms
   *   Comma-separated list of terms.
   * @param array $bundle_fields
   *   Info about a bundle's field instances, as furnished
   *   by field_info_instances().
   *
   * @return object
   *   The query.
   */
  protected function taxonomy($query, $terms, $bundle_fields) {

    $terms = trim($terms);
    if (!empty($terms)) {
      $terms = explode(',', $terms);
      foreach ($bundle_fields as $tname => $value) {
        if ('taxonomy' == $value['widget']['module']) {
          /* The $tname is the taxonomy field name.
           * This works like an OR query.  If the item has been assigned
           * to any of the specified terms, it passes the condition.
           */
          $query->fieldCondition($tname, 'tid', $terms);
          /* We break after we find one term reference field in any of the
           * bundles. Thus, this method only supports cases where we're ok with
           * ignoring any other term reference fields besides the first one
           * in $bundle_fields.  We expect all bundles to have whatever is the
           * first term reference field we find, but no errors result if a
           * bundle doesn't have a term reference field.
           */
          break;
        }
      }
    }

    return $query;
  }

  /**
   * Add the node_view_permissions integration.
   *
   * @param array $bundles
   *   Bundles to check.
   *
   * @return array
   *   Bundles the user is allowed to view.
   */
  protected function nodeViewPermissions($bundles) {

    if (module_exists('node_view_permissions')) {
      $nvp_bundles = node_view_permissions_get_configured_types();
      foreach ($nvp_bundles as $key => $bundle) {
        $perm = 'view any ' . $bundle . ' content';
        if (!user_access($perm)) {
          $key = array_search($bundle, $bundles);
          unset($bundles[$key]);
        };
      }
    }
    return $bundles;

  }

  /**
   * Prevent infinite loops.
   *
   * Prevent display of Entity List fields in list items.
   *
   * @param array $entities
   *   The entities in the list.
   * @param array $config
   *   The list config as defined in the widget form.
   * @param array $bundle_fields
   *   Info about a bundle's field instances, as furnished
   *   by field_info_instances().
   *
   * @return array
   *   Entities.
   */
  protected function safeLoop($entities, $config, $bundle_fields) {

    $has = FALSE;

    extract($config);

    foreach ($bundle_fields as $value) {
      // Check if any of the bundles have an Entity List field.
      if ('entity_list_field' == $value['widget']['module']) {
        $fn = $value['field_name'];
        $has[$fn] = $fn;
      }
    }

    if ($has) {
      foreach ($has as $name) {
        // We need to remove the list fields from the items.
        foreach ($entities as $key => $entity) {
          if (isset($entity->{$name})) {
            unset($entity->{$name});
          }
          $entities[$key] = $entity;
        }
      }
    }

    return $entities;

  }

  /**
   * Helper function for adding a class to the allowed classes array.
   *
   * Modules extending EntityLister should add their class to the allowed
   * classes array, like so:
   *
   * EntityLister::addAllowedClass('MyClass');
   *
   * A good place to do this is in your hook_enable() call in your install file.
   *
   * @param string $class
   *   The class to allow.
   */
  public function addAllowedClass($class) {
    $allowed_classes = variable_get('entity_lister_allowed_classes', array());
    $allowed_classes[$class] = $class;
    variable_set('entity_lister_allowed_classes', $allowed_classes);
  }

  /**
   * Helper function for removing a class from the allowed classes array.
   *
   * Modules extending EntityLister should remove their class in hook_uninstall
   * like so:
   *
   * EntityLister::removeAllowedClass('MyClass');
   *
   * @param string $class
   *   The class to remove.
   */
  public function removeAllowedClass($class) {
    $allowed_classes = variable_get('entity_lister_allowed_classes', array());
    if (isset($allowed_classes[$class])) {
      unset($allowed_classes[$class]);
    }
    variable_set('entity_lister_allowed_classes', $allowed_classes);
  }

}
